from mext.threads.generic.locks import SimpleLock, TimeoutLock

__all__ = ['SimpleEvent', 'MultiEvent', 'LockedEvent', 'Flag']


def _repr(self):
    return "%s()" % self.__class__.__name__

class SimpleEvent(object):
    def __init__(self):
        self.lock = TimeoutLock()
        self.lock.acquire()

    __repr__ = _repr

    def wait(self, timeout=None):
        if self.lock.acquire(timeout):
            self.lock.release()
            return True
        else:
            return False

    def notify(self):
        self.lock.release()
        self.lock.acquire()


class MultiEvent(object):
    def __init__(self):
        self.lock = SimpleLock()
        self.handlers = []

    def __repr__(self):
        return "<%s(%d)>" % (self.__class__.__name__, len(self.handlers))

    def wait(self, timeout=None):
        # None or 0 can be served by simple Lock
        hlock = TimeoutLock() if timeout else SimpleLock()
        hlock.acquire()
        self.handlers.append(hlock)
        if not timeout:
            return hlock.acquire(timeout is None)
        elif hlock.acquire(timeout):
            return True
        with self.lock:
            if hlock in self.handlers:
                self.handlers.remove(hlock)
                return False
            else:
                # this happens if this handler was notified after it timed out
                # but before we got self.lock, so it's the same as if we got it
                return True


    def notify(self, n):
        # pass None to notify all
        with self.lock:
            for hlock in self.handlers[:n]:
                hlock.release()
            self.handlers[:n] = []


class LockedEvent(object):
    """
        A primitive to synchronize changes and threads checking / waiting for those
        changes.

            evt = LockedEvent(lock)
            with lock:
                while not expected state:
                    evt.wait() # wait for relevant state changes

            with lock:
                change state
                evt.notify(1)
    """
    def __init__(self, lock):
        self.lock = lock
        self.handlers = []

    __repr__ = _repr

    def wait(self, timeout=None):
        # here we rely on generic locks not being reentrant
        # (acquiring twice should block until released from another thread)
        # None or 0 can be served by simple Lock
        hlock = TimeoutLock() if timeout else SimpleLock()
        hlock.acquire()
        self.handlers.append(hlock)
        self.lock.release()
        try:
            if not timeout:
                return hlock.acquire(timeout is None)
            elif hlock.acquire(timeout):
                return True
        finally:
            self.lock.acquire()
        if hlock in self.handlers:
            self.handlers.remove(hlock)
            return False
        else:
            # this happens if this handler was notified after it timed out
            # but before we got self.lock, so it's the same as if we got it
            return True

    def notify(self, n):
        # pass None to notify all
        for hlock in self.handlers[:n]:
            hlock.release()
        self.handlers[:n] = []


class Flag(object):
    def __init__(self):
        self.lock = SimpleLock()
        self.evt_set = LockedEvent(self.lock)
        self.flag = False

    def __nonzero__(self):
        return self.flag

    def set(self):
        with self.lock:
            assert not self.flag, "Flag is already set"
            self.flag = True
            self.evt_set.notify(None)

    def clear(self):
        with self.lock:
            self.flag = False

    def wait(self, timeout=None):
        if self.flag:
            return True
        with self.lock:
            if self.flag:
                return True
            self.evt_set.wait(timeout)

